/*
 *  asteroids.c
 *  My very first bona-fide C program
 *
 *  Created by Allen Smith on Oct. 6, 2002.
 *
 *  Scoring System:
 *		Fire a shot:			-levelNumber
 *		Destroy Big Asteroid:	(1+playerMaxStrength-playerStrength) * levelNumber
 *		Destroy Med Asteroid:	(2+playerMaxStrength-playerStrength) * levelNumber
 *		Destroy Small Asteroid:	(3+playerMaxStrength-playerStrength) * levelNumber
 *		Clear Level:			24 * levelNumber
 *		Destroy Opponent:		(144+playerMaxStrength-playerStrength) * levelNumber
 *		Use reverse drive:		-(36 + levelNumber) per 30th of a second
 *
 */
#include "asteroids.h"
//get other headers too
#include "Title Screen.h"
#include "Configure Screen.h"
#include "sound.h"

/*****Global Variables*****/
extern SDL_AudioCVT soundBuffer[];
extern SoundSample channels[NUMBER_CHANNELS];
unsigned int levelNumber = 0; //0 indicates that the game has not yet started
boolean startingNewLevel = false; //this is a flag used while beginning a new level during gameplay. It must be set to false while initializing Level 1
NodePtr shotList = NULL; //a linked list of shots fired
NodePtr asteroidList = NULL; //a linked list of all the asteroids
NodePtr kerblooieList = NULL; //a linked list of explosion particles
NodePtr crystalList = NULL; //a linked list of special crystals
int starfield[NUMBERSTARS][4]; //one row: positionx, positiony, size, opacity
int millisecondsPerStandardFrame = 1000/FRAMES_PER_SECOND;
GLubyte player1Colors[] = {  0, 255,   0}; 
GLubyte player2Colors[] = {  0, 128, 255}; //Aqua

//Global Constants which are important
int asteroidSizes[3] = {10, 20, 30};
int kerblooieSizes[4] = {36, 48, 60, 72};
int *crystalReleaseSchedule[CRYSTAL_NUMBER_OF_SPECIMENS]; //how many crystals should be created; arranged into subarrays indicating the release schedule (see startLevel())
int crystalsReleased[CRYSTAL_NUMBER_OF_SPECIMENS]; //how many crystals have been created (the index of the subarrays above)
int *asteroidReleaseSchedule; //how many medium-sized asteroids to release
int unsmallAsteroidsDestroyedToDate; //how many unsmall asteroids have been hit so far this level
int bigAsteroidsDestroyedToDate; //how many big asteroids have been hit so far this level

boolean paused; //true if the game is paused


//Players
short numberOfPlayers = 0;
short numberOfPlayersActive = 0;
Spaceship player[MAX_NUMBER_PLAYERS];
extern int playerKeys[MAX_NUMBER_PLAYERS][5]; //defined in 'Configure Screen.c'
int playerLives[MAX_NUMBER_PLAYERS] = {0,0};
int playerMaxStrength[MAX_NUMBER_PLAYERS] = {0,0};
boolean playerIsEnabled[MAX_NUMBER_PLAYERS] = {false,false};
int playerScore[MAX_NUMBER_PLAYERS] = {0,0};
int playerMaxLevel[MAX_NUMBER_PLAYERS] = {0,0}; //the highest level attained


/****************************************************
 * gameInit
 *  set up the game and window
 **************************/
void gameInit(int numberOfPlayersChosen){
	unsigned int counter;
	
	//set up glut function calls
	glutMouseFunc(NULL);
	glutKeyboardFunc(handleKeyboard);
	glutKeyboardUpFunc(handleKeyboardUp);
	glutSpecialFunc(handleSpecialKeyboard);
	glutSpecialUpFunc(handleSpecialKeyboardUp);
	glutIgnoreKeyRepeat(true);
	glutDisplayFunc(display);
	glutIdleFunc(idle);

	glPolygonMode(GL_FRONT, GL_FILL);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glClearColor(0,0,0, 1);
	glClear(GL_COLOR_BUFFER_BIT);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	glLoadIdentity();
//	gluOrtho2D(0, FRAMEWIDTH, 0, FRAMEHEIGHT);
	glOrtho(0,FRAMEWIDTH, 0, FRAMEHEIGHT,-640,640);
	
	soundInit();
	
	numberOfPlayers = numberOfPlayersChosen;
	numberOfPlayersActive = numberOfPlayersChosen;
	
	for(counter = 1; counter <= numberOfPlayers; counter++){
		playerScore[P(counter)] = 0;
		playerMaxLevel[P(counter)] = 0; //the highest level attained
		
		//Initialize spaceships
		playerMaxStrength[P(counter)] = 8;
		playerLives[P(counter)] = 3;
		playerIsEnabled[P(counter)] = true;
		player[P(counter)] = spawnSpaceship(counter);
		
	}
	
	//Create blank lists
	shotList = createList();
	asteroidList = createList();
	kerblooieList = createList();
	crystalList = createList();
	
	//start level 1
	#ifdef RANDOM_ON
		srand( time(NULL) );
	#endif
	spawnStarfield();
	startLevel(1);
}

/****************************************************
 * display
 *  paints all the game elements
 **************************/
void display(void) {
	char pausedText[] = "PAUSED"; //print this if the game is paused
	char *levelText;
	unsigned int counter;
	int stringWidth = 0;
	float scaleFactor = 0.75; //75% of original size
		
	drawStarfield();
	
	if(playerIsEnabled[P(1)] == true)
		drawSpaceShip(&player[P(1)]);
	if(playerIsEnabled[P(2)] == true)
		drawSpaceShip(&player[P(2)]);
	
	if(!isEmpty(shotList))
		drawShots(shotList);
	if(!isEmpty(asteroidList))
		drawAsteroids(asteroidList);
	if( !isEmpty(kerblooieList) )
		drawKerblooies(kerblooieList);
	if( !isEmpty(crystalList) )
		drawCrystals(crystalList);
	
	drawStatisticsHeader();

/****	SOME VERY INTERESTING TESTS. IF SIMPLY gluOrtho2d() IN INIT WITH OUT glLoadIdentity() THESE HUGE SHAPES ARE FULLY VISIBLE!
	glColor3ub(255,255,255);
	glBegin(GL_POLYGON);
		glVertex2i(-9000, -9000);
		glVertex2i(-9000,  9000);
		glVertex2i( 9000,  9000);
		glVertex2i( 9000, -9000);
	glEnd();
	
	glColor3ub(255,0,0);
	glLineWidth(1.0);
	glBegin(GL_LINES);
		glVertex2i(  1000,  1000);
		glVertex2i( 90000, 90000);
	glEnd();
****/	
	if(paused == true){ //an overlay to indicate a paused game
		glColor4ub(128,128,128,128); //transparent grey
		glBegin(GL_POLYGON);
			glVertex2i( 0, 0);
			glVertex2i( FRAMEWIDTH, 0);
			glVertex2i( FRAMEWIDTH,  FRAMEHEIGHT);
			glVertex2i(0, FRAMEHEIGHT);
		glEnd();
		
		//Now print the word "PAUSED"
		glColor4ub(255,0,0,192);
		glLineWidth(4.0);
		glPushMatrix();
		for(counter = 0; counter < strlen(pausedText); counter++){
			stringWidth += glutStrokeWidth( GLUT_STROKE_ROMAN , pausedText[counter]);
		}
		glTranslatef( FRAMEWIDTH/2 - (stringWidth/2)*scaleFactor, FRAMEHEIGHT/2, 0);
		glScalef(scaleFactor, scaleFactor, 0);
		glRasterPos2i(0,0);
		for(counter = 0; counter < strlen(pausedText); counter++){
			glutStrokeCharacter( GLUT_STROKE_ROMAN , pausedText[counter]);
		}
		glPopMatrix();
	} //end if(paused)
	
	//Draw big "Level X" if appropriate
	if(startingNewLevel == true && paused == false){ //"Between" levels
		//Print the new level number
		levelText = calloc( strlen( "LEVEL ")+levelNumber+1, sizeof(char) );
		sprintf(levelText, "LEVEL %d", levelNumber+1);
		glColor4ub(128, 0, 64,192); //Maroon with opacity
		glLineWidth(4.0);
		glPushMatrix();
		stringWidth = 0;
		for(counter = 0; counter < strlen(levelText); counter++){
			stringWidth += glutStrokeWidth( GLUT_STROKE_ROMAN , levelText[counter]);
		}
		glTranslatef( FRAMEWIDTH/2 - (stringWidth/2)*scaleFactor, FRAMEHEIGHT/2, 0);
		glScalef(scaleFactor, scaleFactor, 0);
		glRasterPos2i(0,0);
		for(counter = 0; counter < strlen(levelText); counter++){
			glutStrokeCharacter( GLUT_STROKE_ROMAN , levelText[counter]);
		}
		glPopMatrix();
	}
	
	glutSwapBuffers();

}

/****************************************************
 * drawStarfield
 *  draws the stars
 **************************/
void drawStarfield(){
	unsigned int counter;
	
	glClear(GL_COLOR_BUFFER_BIT);
	glBegin(GL_POINTS);
		for(counter = 0; counter < NUMBERSTARS; counter++){
			//one row: positionx, positiony, size, opacity
			glColor4ub(255,255,255, starfield[counter][3]);
			//Note: due to the way in which points are constructed, the values for point size are meaningless. If the points were reimplemented as quadrilaterals, this value would actually be important.
		//	glVertex2iv(starfield[counter]); //Since positions [0] and [1] contain coordinates x,y
		//	Strange! OpenGL Profiler indicates that glVertex2iv takes almost twice as long as glVertex2i to execute!	
			glVertex2i(starfield[counter][0], starfield[counter][1]);
		}//end for
	glEnd();
}


/****************************************************
 * drawStatisticsHeader
 *  Spits out score data and vital signs at the top of the window
 **************************/
void drawStatisticsHeader(){
	char hudText[64]; //"heads-up display?"
	const int hudWidth = 80;
	const int hudHeight = 12;
	const int hudOffsetX = 8;
	const int hudOffsetY = 8;
	int characterWidth; //width of one glut monospaced character
	unsigned int counter;
	float overloadFactor;
	
	//create the status bar for ship strength
	if(	player[P(1)].spaceship->strength > playerMaxStrength[P(1)] ){
		overloadFactor = (float)( player[P(1)].spaceship->strength % playerMaxStrength[P(1)] ) / (playerMaxStrength[P(1)]*OVERLOAD_FACTOR);
		glColor4ub(255*overloadFactor, 255*(1-overloadFactor), 0, 128); //Overload Red
	}
	else if(player[P(1)].spaceship->strength > playerMaxStrength[P(1)]*3/4)
		glColor4ub(0, 255, 0, 128); //Green
	else if(player[P(1)].spaceship->strength > playerMaxStrength[P(1)]*1/2)
		glColor4ub(128, 255, 0, 128); //Lime
	else if(player[P(1)].spaceship->strength > playerMaxStrength[P(1)]*1/4)
		glColor4ub(255, 255, 0, 128); //Yellow
	else // player[P(1)].spaceship->strength > 0
		glColor4ub(255, 0, 0, 128); //Red
	glBegin(GL_POLYGON);
		glVertex2i(hudOffsetX, FRAMEHEIGHT - hudOffsetY	 );
		glVertex2i(hudOffsetX, FRAMEHEIGHT - hudOffsetY - hudHeight);
		glVertex2i(hudOffsetX +player[P(1)].spaceship->strength * hudWidth/playerMaxStrength[P(1)], FRAMEHEIGHT - hudOffsetY - hudHeight);
		glVertex2i(hudOffsetX +player[P(1)].spaceship->strength * hudWidth/playerMaxStrength[P(1)], FRAMEHEIGHT - hudOffsetY );
	glEnd();

	//Create an outline around the status bar, which identifies it by color.
	glColor3ubv(player[P(1)].spaceship->color);
	glLineWidth(2.0);
	glBegin(GL_LINE_LOOP);
		glVertex2i(hudOffsetX, FRAMEHEIGHT - hudOffsetY	 );
		glVertex2i(hudOffsetX, FRAMEHEIGHT - hudOffsetY - hudHeight);
		glVertex2i(hudOffsetX + hudWidth, FRAMEHEIGHT - hudOffsetY - hudHeight);
		glVertex2i(hudOffsetX + hudWidth, FRAMEHEIGHT - hudOffsetY	 );
	glEnd();
	
	//print player 1 vital signs
	sprintf(hudText, "Lives:%2d Score:%6d", playerLives[P(1)] >= 0 ? playerLives[P(1)] : 0, playerScore[P(1)]); //The evil conditional operator: if lives >= 0, print lives, else print '0'
	glRasterPos2i( 2*hudOffsetX + hudWidth, FRAMEHEIGHT - hudHeight - hudOffsetY);
	for(counter = 0; counter < strlen(hudText); counter++)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, hudText[counter]);
	
	//print level number
	glColor3ub(128, 0, 64); //Maroon
	sprintf(hudText, "LEVEL%3d", levelNumber);
	characterWidth = glutBitmapWidth(GLUT_BITMAP_8_BY_13, 'A');
	glRasterPos2i( FRAMEWIDTH/2 - strlen(hudText)*characterWidth/2, FRAMEHEIGHT - hudHeight - hudOffsetY);
	for(counter = 0; counter < strlen(hudText); counter++)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, hudText[counter]);
		
	//Player 2
	if(numberOfPlayers == 2){
		//create the status bar for ship strength. It changes color based on health
		if(	player[P(2)].spaceship->strength > playerMaxStrength[P(2)] ){
			overloadFactor = (float)( player[P(2)].spaceship->strength % playerMaxStrength[P(2)] ) / (playerMaxStrength[P(2)]*OVERLOAD_FACTOR);
			glColor4ub(255*overloadFactor, 255*(1-overloadFactor), 0, 128); //Overload Red
		}
		else if(player[P(2)].spaceship->strength > playerMaxStrength[P(2)]*3/4)
			glColor4ub(0, 255, 0, 128); //Green
		else if(player[P(2)].spaceship->strength > playerMaxStrength[P(2)]*1/2)
			glColor4ub(128, 255, 0, 128); //Lime
		else if(player[P(2)].spaceship->strength > playerMaxStrength[P(2)]*1/4)
			glColor4ub(255, 255, 0, 128); //Yellow
		else // player[P(2)].spaceship->strength > 0
			glColor4ub(255, 0, 0, 128); //Red
		glBegin(GL_POLYGON);
			glVertex2i(FRAMEWIDTH - hudOffsetX -player[P(2)].spaceship->strength * hudWidth/playerMaxStrength[P(2)], FRAMEHEIGHT - hudOffsetY - hudHeight);
			glVertex2i(FRAMEWIDTH - hudOffsetX, FRAMEHEIGHT - hudOffsetY - hudHeight);
			glVertex2i(FRAMEWIDTH - hudOffsetX, FRAMEHEIGHT - hudOffsetY	 );
			glVertex2i(FRAMEWIDTH - hudOffsetX -player[P(2)].spaceship->strength * hudWidth/playerMaxStrength[P(2)], FRAMEHEIGHT - hudOffsetY );
		glEnd();
	
		//Create an outline around the status bar, which identifies it by color.
		glColor3ubv(player[P(2)].spaceship->color);
		glLineWidth(2.0);
		glBegin(GL_LINE_LOOP);
			glVertex2i(FRAMEWIDTH - hudOffsetX, FRAMEHEIGHT - hudOffsetY	 );
			glVertex2i(FRAMEWIDTH - hudOffsetX, FRAMEHEIGHT - hudOffsetY - hudHeight);
			glVertex2i(FRAMEWIDTH - hudOffsetX - hudWidth, FRAMEHEIGHT - hudOffsetY - hudHeight);
			glVertex2i(FRAMEWIDTH - hudOffsetX - hudWidth, FRAMEHEIGHT - hudOffsetY	 );
		glEnd();
		
		//print player 2 vital signs
		sprintf(hudText, "Score:%6d Lives:%2d", playerScore[P(2)], playerLives[P(2)] >= 0 ? playerLives[P(2)] : 0); //The evil conditional operator: if lives >= 0, print lives, else print '0'
		glRasterPos2i( FRAMEWIDTH - (2*hudOffsetX + hudWidth) - strlen(hudText)*characterWidth, FRAMEHEIGHT - hudHeight - hudOffsetY);
		for(counter = 0; counter < strlen(hudText); counter++)
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, hudText[counter]);	
	}//endif for player2
	
}


/****************************************************
 * drawSpaceShip
 *  draws a space ship. Profound name, huh?
 *  actually, it also shows rocket thrust and shield bubbles
 **************************/
void drawSpaceShip(Spaceship * shipPtr){
	float angle; //for creating a circle for shields
	int shipSize = shipPtr->spaceship->shipSize;
	int spriteTimeRemaining; //how much longer fading effects like shields and thrust have left
	float spriteAlpha; //the *percentage* alpha (transparency) for the thrust and shields
	int currentTime = glutGet( GLUT_ELAPSED_TIME );
	boolean shieldsOverloading;
	
	glPushMatrix();
	glTranslatef(shipPtr->position.x, shipPtr->position.y, 0);
	glRotatef(shipPtr->spaceship->rotationAngle, 0,0,1);

	/* Basic Ship Model:
		glVertex2i( x, y+20);
		glVertex2i( x-10, y-10*sqrt(3) );
		glVertex2i( x, y-15);
		glVertex2i( x+10, y-10*sqrt(3) );
	*/
	
	//rocket thrust; drawn first so it appears behind the ship.
	spriteTimeRemaining = shipPtr->spaceship->drawThrustUntil - currentTime;
	if(shipPtr->spaceship->thrusting == true || spriteTimeRemaining > 0){
		if(shipPtr->spaceship->thrusting == true)
			spriteAlpha = 1.0; //100%
		else
			spriteAlpha = spriteTimeRemaining / (float) FADEOUTS;
		
		glBegin(GL_POLYGON);
			if(shipPtr->spaceship->strength <= playerMaxStrength[P(shipPtr->spaceship->playerNumber)])
				glColor4ub( 255 * shipPtr->spaceship->strength/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 0, 0, 255*spriteAlpha);
			else //overloading
				glColor4ub( 255, 255*(shipPtr->spaceship->strength%playerMaxStrength[P(shipPtr->spaceship->playerNumber)])/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 255*(shipPtr->spaceship->strength%playerMaxStrength[P(shipPtr->spaceship->playerNumber)])/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 255*spriteAlpha);
			glVertex2i( 0, 0); //centerpoint; to complete the filled-in appearance
			
			glColor4ub( 128, 0, 0, 255*spriteAlpha);
			glVertex2f( (shipSize*4/5)*cos(5*PI/6), //port blast
						(shipSize*4/5)*sin(5*PI/6));
						
			if(shipPtr->spaceship->strength <= playerMaxStrength[P(shipPtr->spaceship->playerNumber)])
				glColor4ub( 255 * shipPtr->spaceship->strength/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 0, 0, 255*spriteAlpha);
			else //overloading
				glColor4ub( 255, 255*(shipPtr->spaceship->strength%playerMaxStrength[P(shipPtr->spaceship->playerNumber)])/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 255*(shipPtr->spaceship->strength%playerMaxStrength[P(shipPtr->spaceship->playerNumber)])/playerMaxStrength[P(shipPtr->spaceship->playerNumber)], 255*spriteAlpha);
			glVertex2f( (shipSize*1.5)*cos(1*PI/1), //center blast
						(shipSize*1.5)*sin(1*PI/1));
						
			glColor4ub( 128, 0, 0, 255*spriteAlpha);
			glVertex2f( (shipSize*4/5)*cos(7*PI/6), //starboard blast
						(shipSize*4/5)*sin(7*PI/6));
						
		glEnd();
	}//end if

	//the spaceship
	glColor3ubv(shipPtr->spaceship->color);
	glBegin(GL_POLYGON);
		glVertex2f( shipSize, //nose point
					0);
					
		glVertex2f( shipSize*cos(5*PI/6), //port fin
					shipSize*sin(5*PI/6));
					
		glVertex2f( (shipSize*3/5)*cos(1*PI/1), //center point
					(shipSize*3/5)*sin(1*PI/1));
					
		glVertex2f( shipSize*cos(7*PI/6), //starboard fin
					shipSize*sin(7*PI/6));
	glEnd();
	
	//A Border for the ship
	glColor3ub(shipPtr->spaceship->color[0]*0.25, shipPtr->spaceship->color[1]*0.25, shipPtr->spaceship->color[2]*0.25); //darker color
	glLineWidth(2.0);
	glBegin(GL_LINE_LOOP);
		glVertex2f( shipSize, //nose point
					0);
					
		glVertex2f( shipSize*cos(5*PI/6), //port fin
					shipSize*sin(5*PI/6));
					
		glVertex2f( (shipSize*3/5)*cos(1*PI/1), //center point
					(shipSize*3/5)*sin(1*PI/1));
					
		glVertex2f( shipSize*cos(7*PI/6), //starboard fin
					shipSize*sin(7*PI/6));
					
	glEnd();
	
	//Draw Shields
	if(player[P(shipPtr->spaceship->playerNumber)].spaceship->strength > playerMaxStrength[P(shipPtr->spaceship->playerNumber)])
		shieldsOverloading = true;
	else
		shieldsOverloading = false;
	if(shipPtr->spaceship->drawShieldsTimeToLive > 0 || shieldsOverloading){
		if(shipPtr->spaceship->invinciblityTimeToLive > 0 )
			spriteAlpha = 0.3125; // 5/16ths
		else if(shieldsOverloading)
			spriteAlpha = (float)( player[P(shipPtr->spaceship->playerNumber)].spaceship->strength % playerMaxStrength[P(shipPtr->spaceship->playerNumber)] ) / (playerMaxStrength[P(shipPtr->spaceship->playerNumber)]*OVERLOAD_FACTOR);
		else
			spriteAlpha = shipPtr->spaceship->drawShieldsTimeToLive / (float) FADEOUTS;
		glLineWidth(2.0);
		if(shieldsOverloading)
			glColor4ub(	255*spriteAlpha, 0, 0, 128*spriteAlpha);
		else
			glColor4ub(	0, 128, 255, 128*spriteAlpha);
		glBegin(GL_LINE_LOOP);
			for(angle = 0; angle < 2*PI; angle += 2*PI/12){ //twelve-sided shield bubble
				glVertex2f( (shipSize*0.75 +1)*cos(angle), (shipSize*0.75 +1)*sin(angle) );
			}
		glEnd();
		
		//Central Shild Line!
		glLineWidth(3.0);
		if(shieldsOverloading)
			glColor4ub(	255*spriteAlpha, 0, 0, 255*spriteAlpha);
		else
			glColor4ub(	0, 128, 255, 255*spriteAlpha);
		glBegin(GL_LINE_LOOP);
			for(angle = 0; angle < 2*PI; angle += 2*PI/12){
				glVertex2f( (shipSize*0.75)*cos(angle), (shipSize*0.75)*sin(angle) );
			}
		glEnd();
		
		glLineWidth(2.0);
		if(shieldsOverloading)
			glColor4ub(	255*spriteAlpha, 0, 0, 128*spriteAlpha);
		else
			glColor4ub(	0, 128, 255, 128*spriteAlpha);
		glBegin(GL_LINE_LOOP);
			for(angle = 0; angle < 2*PI; angle += 2*PI/12){
				glVertex2f( (shipSize*0.75 -1)*cos(angle), (shipSize*0.75 -1)*sin(angle) );
			}
		glEnd();

	}

	glPopMatrix();
}


/****************************************************
 * drawShots
 *  accepts a linked list of shots and paints them
 **************************/
void drawShots(NodePtr shots){
	Object shotToDraw;
	NodePtr currentPtr;
	float angle;
	currentPtr = shots;
	
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		do{
			shotToDraw = currentPtr->object;
			glColor3ubv( shotToDraw.shot.color );
			
			glPushMatrix();
			glTranslatef(shotToDraw.position.x, shotToDraw.position.y, 0);
			//one could just draw points for these. However, drawing circles looks better, especially on resize
			glBegin(GL_TRIANGLE_FAN);
				for(angle = 0; angle < 2*PI; angle += 2*PI/12){
					glVertex2f( 2*cos(angle), 
								2*sin(angle) );
				}
			glEnd();
			glPopMatrix();
			
			currentPtr = currentPtr->nextNode;
		}while( currentPtr != NULL && currentPtr->nextNode != NULL);
	}
}


/****************************************************
 * drawAsteroids
 *  accepts a linked list of Asteroids and paints them
 **************************/
void drawAsteroids(NodePtr asteroids){
	Asteroid asteroidToDraw;
	NodePtr currentPtr;
	float angle;
	static int numberVertices = 7;
	int size;
	
	currentPtr = asteroids;
	if(currentPtr->previousNode == NULL) //at head
		currentPtr = currentPtr->nextNode;
	if(currentPtr != NULL){
		do{
			glPushMatrix();
			asteroidToDraw = currentPtr->object;
			size = asteroidSizes[asteroidToDraw.asteroid.size]; //lookup the value
			glColor3ubv(asteroidToDraw.asteroid.color);
			
			glTranslatef(asteroidToDraw.position.x, asteroidToDraw.position.y, 0);
			glBegin(GL_TRIANGLE_FAN);
				for(angle = 0; angle < 2*PI; angle += 2*PI/numberVertices){
					glVertex2f( size*cos(angle), 
								size*sin(angle) );
				}
			glEnd();
			
			currentPtr = currentPtr->nextNode;
			glPopMatrix();
		}while( currentPtr != NULL && currentPtr->nextNode != NULL);
	}
}


/****************************************************
 * drawKerblooies
 *  accepts a linked list of kerblooies and paints them
 **************************/
void drawKerblooies(NodePtr kerblooies){
	float angle; //used to draw round particle
	Object kerblooieToDraw;
	NodePtr currentPtr;
	float randomParticleSize; //each kerblooie particle's size is set to "twinkle" slightly
	
	currentPtr = kerblooies;
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		do{
			kerblooieToDraw = currentPtr->object;
			
			glPushMatrix();
			glTranslatef(kerblooieToDraw.position.x, kerblooieToDraw.position.y, 0);
			glColor4ubv( kerblooieToDraw.kerblooie.color );
			randomParticleSize = kerblooieToDraw.kerblooie.size + rand()%2 * 0.25;
			glBegin(GL_TRIANGLE_FAN);
				for(angle = 0; angle < 2*PI; angle += 2*PI/5){
					glVertex2f( randomParticleSize/2*cos(angle), 
								randomParticleSize/2*sin(angle) );
				}
			glEnd();
			glPopMatrix();

			currentPtr = currentPtr->nextNode;
		}while( currentPtr != NULL && currentPtr->nextNode != NULL);
	}
}


/****************************************************
 * drawCrystals
 *  accepts a linked list of crystals and paints them
 **************************/
void drawCrystals(NodePtr crystals){
	Object crystalToDraw;
	NodePtr currentPtr = crystals;
	short counter;
	static short armVectors[3][3] = {	{1,0,0},
										{0,1,0},
										{0,0,1}
									};
	
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		glLineWidth(1);
		do{
			crystalToDraw = currentPtr->object;
			glPushMatrix();
			glTranslatef(crystalToDraw.position.x, crystalToDraw.position.y, 0);
			glRotatef(crystalToDraw.crystal->rotationAngle, 0,1,1);
			
			for(counter = 0; counter < 3; counter++){ //draw the three struts/arms of the crystal
				glPushMatrix();
				glRotatef(90,	armVectors[counter][0],
								armVectors[counter][1],
								armVectors[counter][2]  );
				glColor4ubv( crystalToDraw.crystal->color );
				drawCrystalArm(crystalToDraw.crystal->radius, crystalToDraw.crystal->thickness);
				glColor4ub(	crystalToDraw.crystal->color[0],
							crystalToDraw.crystal->color[1],
							crystalToDraw.crystal->color[2],
							crystalToDraw.crystal->color[2]/2 );
				drawCrystalArm(	crystalToDraw.crystal->radius*(1 + 3/32),
								crystalToDraw.crystal->thickness + crystalToDraw.crystal->radius*3/32);
				glPopMatrix();	
			}//end for to draw arms
			glPopMatrix();
			currentPtr = currentPtr->nextNode;
		}while( currentPtr != NULL && currentPtr->nextNode != NULL);
	}
}

/****************************************************
 * drawCrystalArm
 *  generates a single arm for a crystal
 **************************/
void drawCrystalArm(float radius, float thickness){
	float angle;
	static int slices = 6;
	glBegin(GL_QUAD_STRIP);
		for(angle = 0; angle <= 2*PI; angle += 2*PI/slices){
			glVertex3f( thickness * cos(angle), 
						thickness * sin(angle),
						-radius );
			glVertex3f( thickness * cos(angle), 
						thickness * sin(angle),
						+radius );
		}
	glEnd();
	
	glPushMatrix();
	glTranslatef(0,0,-radius);
	glRotatef(180,1,0,0);
	drawHalfSphere(thickness,slices,5);
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(0,0,+radius);
	drawHalfSphere(thickness,slices,5);
	glPopMatrix();
}


/****************************************************
 * drawHalfSphere
 *  oriented based in x-y plane, rising in +z
 **************************/
void drawHalfSphere(int radius, int slices, int stacks){
	float angle;
	int stack;
	float radiusStack;
	for(stack = 0; stack < stacks; stack++){
		glBegin(GL_QUAD_STRIP);
			for(angle = 0; angle <= 2*PI; angle += 2*PI/slices){
				radiusStack = radius*sqrt(1-pow((float)stack/stacks,2));
				glVertex3f( radiusStack * cos(angle), 
							radiusStack * sin(angle),
							(float)(stack)*radius/stacks );
				
				radiusStack = radius*sqrt(1-pow( (float)(stack+1)/stacks, 2) );
				glVertex3f( radiusStack * cos(angle), 
							radiusStack * sin(angle),
							(float)(stack+1)*radius/stacks );
			}
		glEnd();
	}
}


/****************************************************
 * moveSpaceShip
 *  adjust's a ship's location based on its direction vector
 **************************/
void moveSpaceShip(Spaceship * shipPtr, float movementScaleFactor){
	unsigned int counter;
	int numberWarningDingsPlaying = 0;
	
#ifndef ETERNAL_INVINCIBILITY
//If NOT defined
	if(shipPtr->spaceship->invinciblityTimeToLive > 0)
		shipPtr->spaceship->invinciblityTimeToLive -= millisecondsPerStandardFrame * movementScaleFactor;
	if(shipPtr->spaceship->drawShieldsTimeToLive > 0)
		shipPtr->spaceship->drawShieldsTimeToLive -= millisecondsPerStandardFrame * movementScaleFactor;
#endif
	if(shipPtr->spaceship->strength == playerMaxStrength[P(shipPtr->spaceship->playerNumber)]*(1+OVERLOAD_FACTOR) ){
		for(counter = 0; counter < NUMBER_CHANNELS; counter++){
			if(channels[counter].soundID == WARNING_SOUND(shipPtr->spaceship->playerNumber) && 
					channels[counter].dataPosition != channels[counter].dataLength)
				numberWarningDingsPlaying++; //count how many times the thrust sound effect is playing
		}
		if(numberWarningDingsPlaying < 1)
			playSound(soundBuffer, WARNING_SOUND(shipPtr->spaceship->playerNumber) );
	}

	adjustCoordinate(&shipPtr->position, &shipPtr->direction, movementScaleFactor);
}

/****************************************************
 * moveShots
 *  adjust's all shots' locations based on their direction vectors
 **************************/
void moveShots(NodePtr * shots, float movementScaleFactor){
	Shot *shotToMovePtr;
	NodePtr nextPtr;
	NodePtr originalPtr = *shots;
	if((*shots)->previousNode == NULL)
		*shots = (*shots)->nextNode;//advance past head
	if(*shots != NULL && (*shots)->nextNode != NULL){
		do{
			nextPtr = (*shots)->nextNode;
			shotToMovePtr = &((*shots)->object);
			
			shotToMovePtr->shot.timeToLive -= millisecondsPerStandardFrame * movementScaleFactor;
			if(shotToMovePtr->shot.timeToLive > 0 &&
				shotToMovePtr->shot.impacted == false){
				adjustCoordinate( &(shotToMovePtr->position), &(shotToMovePtr->direction), movementScaleFactor );
			}
			else{
				#ifdef DEBUG_SHOTS
					printf("At time %d shot %d > 8\n",time(NULL),shotToMovePtr->shot.timeFired);
				#endif
				removeNode(shots);
			}
			*shots = nextPtr;
		}while( *shots != NULL && (*shots)->nextNode != NULL );
	}
	*shots = originalPtr;
	
}

/****************************************************
 * moveAsteroids
 *  adjust's all asteroids' locations based on their direction vectors
 *  If an asteroid has been hit by a shot, appropriate action is taken
 **************************/
void moveAsteroids(NodePtr * asteroids, float movementScaleFactor){
	Asteroid *asteroidToMovePtr;
	NodePtr nextPtr;
	NodePtr originalPtr = *asteroids;
	
	short asteroidsToRelease;
	int oldAsteroidSize;
	unsigned int counter;
	Asteroid newAsteroid;
	
if(*asteroids != NULL){
	if((*asteroids)->previousNode == NULL)
		*asteroids = (*asteroids)->nextNode;//advance past head
	if((*asteroids)->nextNode != NULL){
		do{
			nextPtr = (*asteroids)->nextNode;
			asteroidToMovePtr = &((*asteroids)->object);
			adjustCoordinate( &(asteroidToMovePtr->position),
							  &(asteroidToMovePtr->direction),
							  movementScaleFactor );
			oldAsteroidSize = asteroidToMovePtr->asteroid.size;
			isAsteroidHit(asteroidToMovePtr, &shotList); //if it is, an impacted flag is activated
			
			if(asteroidToMovePtr->asteroid.impacted.hitCode == true){
				if(oldAsteroidSize > ASTEROID_SMALL){
					//Do we release crystals?
					for(counter = 0; counter < CRYSTAL_NUMBER_OF_SPECIMENS; counter++){
						if(crystalReleaseSchedule[counter][ crystalsReleased[counter] ] == unsmallAsteroidsDestroyedToDate ){
							spawnCrystal(
								counter, //What kind of crystal
								asteroidToMovePtr->position,
								&crystalList );
							crystalsReleased[counter]++;
							#ifdef DEBUG_CRYSTALS
								printf("C%d: %d\n",counter, crystalsReleased[counter]);
							#endif
						}
					}
					
					//Create new, smaller asteroids
					if(oldAsteroidSize == ASTEROID_BIG){
						asteroidsToRelease = asteroidReleaseSchedule[bigAsteroidsDestroyedToDate]; //schedule defined in startLevel()
						//printf("Number New: %d\n",asteroidReleaseSchedule[bigAsteroidsDestroyedToDate]);
						bigAsteroidsDestroyedToDate++;
					}
					else //This will happen on ASTEROID_MEDIUM
						asteroidsToRelease = rand()%2+2; //just release a random number. No need to schedule these.
					for(counter = 0; counter < asteroidsToRelease; counter++){ 
						newAsteroid = spawnAsteroid(
							oldAsteroidSize - 1,
							asteroidToMovePtr->position.x + ( rand()%10 *pow(-1,rand()) ),
							asteroidToMovePtr->position.y + ( rand()%10 *pow(-1,rand()) ),
							asteroidToMovePtr->direction
						);
						insertNode(&originalPtr, newAsteroid);
					}
					#ifdef DEBUG_CRYSTALS
						printf("Impacts: %d\n",unsmallAsteroidsDestroyedToDate);
					#endif
					unsmallAsteroidsDestroyedToDate++; //just a counter for unsmall (potentially crystal-causing) impacts
				}
				spawnKerblooie(
					oldAsteroidSize,
					asteroidSizes[oldAsteroidSize]/2,
					asteroidToMovePtr->asteroid.color,
					asteroidToMovePtr->position,
					asteroidToMovePtr->direction,
					&kerblooieList
				);
				removeNode(asteroids); //since it has been hit by a shot!
			}
			*asteroids = nextPtr;
		}while( *asteroids != NULL && (*asteroids)->nextNode != NULL );
	}
}
	*asteroids = originalPtr;
	
}


/****************************************************
 * moveKerblooies
 *  adjusts all explosion particles appropriately
 **************************/
void moveKerblooies(NodePtr * kerblooies, float movementScaleFactor){
	Kerblooie *kerblooiesToMovePtr;
	NodePtr nextPtr;
	NodePtr originalPtr = *kerblooies;
	float colorFade;
	if((*kerblooies)->previousNode == NULL)
		*kerblooies = (*kerblooies)->nextNode; //advance past head
	if(*kerblooies != NULL && (*kerblooies)->nextNode != NULL){
		do{
			nextPtr = (*kerblooies)->nextNode;
			kerblooiesToMovePtr = &((*kerblooies)->object);
			
			kerblooiesToMovePtr->kerblooie.timeToLive -= millisecondsPerStandardFrame * movementScaleFactor;
			if(kerblooiesToMovePtr->kerblooie.timeToLive > 0){
				adjustCoordinate(	&(kerblooiesToMovePtr->position),
									&(kerblooiesToMovePtr->direction),
									movementScaleFactor );
				colorFade =  kerblooiesToMovePtr->kerblooie.timeToLive / (float) kerblooiesToMovePtr->kerblooie.lifespan;
				kerblooiesToMovePtr->kerblooie.color[3] = floor( 255 * colorFade); //adjust alpha, which starts at 255
			}
			else{
				removeNode(kerblooies);
			}
			*kerblooies = nextPtr;
		}while( *kerblooies != NULL && (*kerblooies)->nextNode != NULL );
	}
	*kerblooies = originalPtr; //restore the list pointer to the beginning
	
}

/****************************************************
 * moveCrystals
 *  adjusts all crystals appropriately, killing those that have expired
 **************************/
void moveCrystals(NodePtr * crystals, float movementScaleFactor){
	Crystal *crystalsToMovePtr;
	NodePtr nextPtr;
	NodePtr originalPtr = *crystals;
	
	if((*crystals)->previousNode == NULL)
		*crystals = (*crystals)->nextNode; //advance past head
	if(*crystals != NULL && (*crystals)->nextNode != NULL){
		do{
			nextPtr = (*crystals)->nextNode;
			crystalsToMovePtr = &((*crystals)->object);
			crystalsToMovePtr->crystal->timeToLive -= millisecondsPerStandardFrame * movementScaleFactor;
			isCrystalHit(crystalsToMovePtr, &shotList); //First determine if the crystal is still alive
			if(crystalsToMovePtr->crystal->timeToLive > 0 && crystalsToMovePtr->crystal->status.hitCode == CRYSTAL_NORMAL){
				adjustCoordinate(	&(crystalsToMovePtr->position),
									&(crystalsToMovePtr->direction),
									movementScaleFactor );
				crystalsToMovePtr->crystal->rotationAngle = fmod(crystalsToMovePtr->crystal->rotationAngle + crystalsToMovePtr->crystal->rotationIncrement * movementScaleFactor, 360);
			}
			else{
				if(crystalsToMovePtr->crystal->status.hitCode == CRYSTAL_RETRIEVED)
					playSound(soundBuffer, RETRIEVE_CRYSTAL_SOUND);
				else{
					if(crystalsToMovePtr->crystal->status.hitCode == CRYSTAL_IMPACTED) //crystal shot
						playSound(soundBuffer, CRYSTAL_SPLATTER_SOUND);
					//else if(crystalsToMovePtr->crystal->timeToLive <= 0)
					spawnKerblooie(
						KERBLOOIE_SMALL,
						crystalsToMovePtr->crystal->radius/2,
						crystalsToMovePtr->crystal->color,
						crystalsToMovePtr->position,
						crystalsToMovePtr->direction,
						&kerblooieList );
				}
				free(crystalsToMovePtr->crystal);
				removeNode(crystals);
			}
			*crystals = nextPtr;
		}while( *crystals != NULL && (*crystals)->nextNode != NULL );
	}
	*crystals = originalPtr; //restore the list pointer to the beginning
	
}

/****************************************************
 * adjustCoordinate
 *	modifies coordPtr by the amount of the directionPtr
 *  wraps a point around the window
 **************************/
void adjustCoordinate(Coordinate *coordPtr, Vector *directionPtr, float movementScaleFactor){
	coordPtr->x += directionPtr->x * movementScaleFactor;
	coordPtr->y += directionPtr->y * movementScaleFactor;
	
	coordPtr->x = fmod( fmod(coordPtr->x,FRAMEWIDTH)+FRAMEWIDTH, FRAMEWIDTH );
	coordPtr->y = fmod( fmod(coordPtr->y,FRAMEHEIGHT)+FRAMEHEIGHT, FRAMEHEIGHT );
}

/****************************************************
 * collideObjects
 *	adds vector1 to vector2 and affects them accordingly
 *	an actual physics equation!
 *	Note: 25% of velocity is lost in a collision.
 **************************/
void collideObjects(Object *hitter, Object *hittee){
	hitter->direction.x = 0.75 * ( (hitter->mass - hittee->mass)*hitter->direction.x + 2 * hittee->mass * hitter->direction.x) / (hitter->mass + hittee->mass);
	hitter->direction.y = 0.75 * ( (hitter->mass - hittee->mass)*hitter->direction.y + 2 * hittee->mass * hitter->direction.y) / (hitter->mass + hittee->mass);
	
	hittee->direction.x = 0.75 * (2 * hitter->mass * hitter->direction.x + (hittee->mass - hitter->mass) * hittee->direction.x) / (hitter->mass + hittee->mass);
	hittee->direction.y = 0.75 * (2 * hitter->mass * hitter->direction.y + (hittee->mass - hitter->mass) * hittee->direction.y) / (hitter->mass + hittee->mass);
}

/****************************************************
 * idle
 *  controls animation
 **************************/
void idle(void){
	static int previousSystemTime;
	unsigned int counter;
	float movementScaleFactor;
	
	
	movementScaleFactor = (float)(glutGet(GLUT_ELAPSED_TIME) - previousSystemTime)/millisecondsPerStandardFrame;
	movementScaleFactor = fmod(movementScaleFactor, 1);
	
	//Record the new "previous system time" before doing all the movement calculations
	previousSystemTime = glutGet(GLUT_ELAPSED_TIME);


	//player operations
	for(counter = 1; counter <= numberOfPlayers; counter++){
		if(playerIsEnabled[P(counter)] == true){
			if(player[P(counter)].spaceship->rotating != false)
				rotateShip(&(player[P(counter)]), movementScaleFactor);
			if(player[P(counter)].spaceship->thrusting == true)
				accelerateShip(&(player[P(counter)]), movementScaleFactor);
			if(player[P(counter)].spaceship->decelerating == true)
				decelerateShip(&(player[P(counter)]), movementScaleFactor);
			
			moveSpaceShip(&(player[P(counter)]), movementScaleFactor);
		}
	}
	
	//generic objects
	if(shotList != NULL)
		moveShots(&shotList, movementScaleFactor);
	if(asteroidList != NULL)
		moveAsteroids(&asteroidList, movementScaleFactor);
	if( !isEmpty(kerblooieList) )
		moveKerblooies(&kerblooieList, movementScaleFactor);
	if( !isEmpty(crystalList) )
		moveCrystals(&crystalList, movementScaleFactor);
		
	//assess damage (must be done after shots are moved)
	for(counter = 1; counter <= numberOfPlayers; counter++){
		if(playerIsEnabled[P(counter)] == true){
			impactShip(&player[P(counter)], isShipHit(&(player[P(counter)]), &asteroidList, &shotList, &crystalList) );
		}
	}
	if(isEmpty(asteroidList) && startingNewLevel == false){
		glutTimerFunc(0, startLevel, levelNumber+1); //start new level after four seconds (handled inside function call)
		startingNewLevel = true;
	}
	
	//windowing
	glutPostRedisplay();
}

/****************************************************
 * fireShot
 *  fires a shot by the designated player
 **************************/
void fireShot(Spaceship * player){
	Shot newshot;
	
	playSound(soundBuffer, POP_SOUND);
		
	newshot.shot.owner = player->spaceship->playerNumber;
	newshot.shot.timeToLive =  1000; //1 second
	newshot.shot.impacted = false;
	
	newshot.position.x = player->position.x +
						player->spaceship->shipSize*cos(player->spaceship->rotationAngle * PI/180 ) *1.125; //move the shot away from the ship
	newshot.position.y = player->position.y +
						player->spaceship->shipSize*sin(player->spaceship->rotationAngle * PI/180 ) *1.125;
						
	newshot.direction.x = player->direction.x +
						player->spaceship->strength* cos(player->spaceship->rotationAngle * PI/180);
	newshot.direction.y = player->direction.y +
						player->spaceship->strength* sin(player->spaceship->rotationAngle * PI/180);
	
	newshot.mass = 1;
	newshot.shot.color[0] = player->spaceship->color[0];
	newshot.shot.color[1] = player->spaceship->color[1];
	newshot.shot.color[2] = player->spaceship->color[2];
	
	playerScore[P(player->spaceship->playerNumber)] -= levelNumber; //it costs you to shoot!
	insertNode(&shotList,newshot);
}

/****************************************************
 * isCrystalHit
 *  returns true if the crystal has been hit by a shot
 *  the shot that hit the crystal is flagged for deletion from the list
 **************************/
HitData isCrystalHit(Crystal *crystal, NodePtr *shots){
	Shot *shotToTest;
	NodePtr currentPtr;
	float distance; //result of distance between point and crystal
	HitData isHit;
	
	isHit.hitCode = false;
	currentPtr = *shots;
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		while( currentPtr != NULL && currentPtr->nextNode != NULL && isHit.hitCode == false){
			shotToTest = &currentPtr->object;
			distance = distanceCoordinate(shotToTest->position, crystal->position);
			if(distance < crystal->crystal->radius){
				isHit.hitCode = CRYSTAL_IMPACTED;
				isHit.mass = shotToTest->mass;
				isHit.direction = currentPtr->object.direction;
				currentPtr->object.shot.impacted = true;
				crystal->crystal->status = isHit;
				collideObjects(shotToTest, crystal);
			}
			else
				currentPtr = currentPtr->nextNode;
		}
	}
	return isHit;
}


/****************************************************
 * isAsteroidHit
 *  returns true if the asteroid has been hit by a shot
 *  the shot that hit the asteroid is flagged for deletion from the list
 **************************/
HitData isAsteroidHit(Asteroid *asteroid, NodePtr *shots){
	Shot *shotToTest;
	NodePtr currentPtr;
	float distance; //result of distance between point and asteroid
	HitData isHit;
	int shotOwner;
	
	isHit.hitCode = false;
	
	currentPtr = *shots;
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		while( currentPtr != NULL && currentPtr->nextNode != NULL && isHit.hitCode == false){
			shotToTest = &currentPtr->object;
			distance = distanceCoordinate(shotToTest->position, asteroid->position);
			if(distance < asteroidSizes[asteroid->asteroid.size]){ //a hit!
				isHit.hitCode = true;
				isHit.mass = shotToTest->mass;
				isHit.direction = currentPtr->object.direction;
				currentPtr->object.shot.impacted = true;
				asteroid->asteroid.impacted = isHit;
				collideObjects( shotToTest, asteroid );
				
				//Scoring:
				shotOwner = shotToTest->shot.owner;
				
				if(player[P(shotOwner)].spaceship->strength > 0)
					playerScore[P(shotOwner)] += 
						( abs(asteroid->asteroid.size - 3) +
							playerMaxStrength[P(shotOwner)] /
							player[P(shotOwner)].spaceship->strength )
						* levelNumber;
				else
					playerScore[P(shotOwner)] += 
						( abs(asteroid->asteroid.size - 3) +
							playerMaxStrength[P(shotOwner)] )
						* levelNumber;
			}
			else
				currentPtr = currentPtr->nextNode;
		}
	}
	return isHit;
}

/****************************************************
 * isShipHit
 *  returns int for amount by which to decrease ship strength
 *   -> 1 for a shot, otherwise dependent on asteroid size
 *  the shot that hit the ship is flagged for deletion from the list
 **************************/
HitData isShipHit(Spaceship *shipPtr, NodePtr *asteroids, NodePtr *shots, NodePtr *crystals){
	Crystal *crystalToTest;
	Shot *shotToTest;
	Asteroid *asteroidToTest;
	NodePtr currentPtr;
	float distance; //result of distance between point and spaceship
	HitData isHit; //how much impact did the ship absorb?
	int shotOwner;
	int playerNumber = shipPtr->spaceship->playerNumber;
	int scoreAssessed;
	
	isHit.hitCode = 0; //Stores the number of impacts sustained; no hits yet
	isHit.mass = 0;
	isHit.direction.x = 0;
	isHit.direction.y = 0;
	
//Test crystals for intersection
	currentPtr = *crystals;
	if(currentPtr != NULL && player[P(playerNumber)].spaceship->invinciblityTimeToLive <= 0){ //only pick up crystals if the ship is not invincible (after all, crystals can sometimes be dangerous!)
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		while( currentPtr != NULL && currentPtr->nextNode != NULL){
			crystalToTest = &currentPtr->object;
			distance = distanceCoordinate(crystalToTest->position, shipPtr->position);
			if(distance < crystalToTest->crystal->radius + shipPtr->spaceship->shipSize*0.75){
				crystalToTest->crystal->status.hitCode = CRYSTAL_RETRIEVED;
				//Assess result of crystal collection
				switch(crystalToTest->crystal->crystalFlavor){
					case CRYSTAL_SHIELDPOWER:
						playSound(soundBuffer, RETRIEVE_CRYSTAL_SOUND);
						player[P(playerNumber)].spaceship->strength++;
					/* Code required to give extra life after exceeding max strength (not used)
						if(player[P(playerNumber)].strength > playerMaxStrength[P(playerNumber)]){
							player[P(playerNumber)].strength %= playerMaxStrength[P(playerNumber)];
							playerLives[P(playerNumber)]++;
						}
					*/	
						if(player[P(playerNumber)].spaceship->strength > playerMaxStrength[P(playerNumber)]*(1+OVERLOAD_FACTOR) )
							isHit.hitCode = player[P(playerNumber)].spaceship->strength; //destroy players who overload their shields
						break;
				}
			}
			currentPtr = currentPtr->nextNode;
		}
	}

//Test shots for impact
	currentPtr = *shots;
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		while( currentPtr != NULL && currentPtr->nextNode != NULL){
			shotToTest = &currentPtr->object;
			distance = distanceCoordinate(shotToTest->position, shipPtr->position);
			if(distance < shipPtr->spaceship->shipSize){
				collideObjects(shotToTest, shipPtr);
				isHit.mass += shotToTest->mass;
				isHit.hitCode++;
				currentPtr->object.shot.impacted = true;
				
				//Scoring:
				shotOwner = shotToTest->shot.owner;
				//Don't award points for shooting oneself.
				//Potential Computer Players might be given negative numbers, hence the second condition
				if(shotOwner != playerNumber && shotOwner >= 1){
					if(player[P(shotOwner)].spaceship->strength > 0)
						playerScore[P(shotOwner)] += 
							(144 +
								playerMaxStrength[P(shotOwner)] / 
								player[P(shotOwner)].spaceship->strength
							) * levelNumber;
					else
						scoreAssessed =
							(144 +
								playerMaxStrength[P(shotOwner)]
							) * levelNumber;
				}

			}
			currentPtr = currentPtr->nextNode;
		}
	}

//Test asteroids for impact
	currentPtr = *asteroids;
	if(currentPtr != NULL){
		if(currentPtr->previousNode == NULL)
			currentPtr = currentPtr->nextNode;
		while( currentPtr != NULL && currentPtr->nextNode != NULL){
			asteroidToTest = &currentPtr->object;
			distance = distanceCoordinate(asteroidToTest->position, shipPtr->position);
			if(distance < asteroidSizes[asteroidToTest->asteroid.size] + shipPtr->spaceship->shipSize*0.75){
				collideObjects(asteroidToTest, shipPtr);
				isHit.mass += asteroidToTest->mass;
				isHit.hitCode += (asteroidToTest->asteroid.size*2 +1);
				
				asteroidToTest->asteroid.impacted.hitCode = true;
				asteroidToTest->asteroid.impacted.mass = shipPtr->mass;
				asteroidToTest->asteroid.impacted.direction = shipPtr->direction;
				
				//Scoring (yes, you get points for ramming the asteroids!):
				if(player[P(playerNumber)].spaceship->strength > 0)
					scoreAssessed = 
						( abs(asteroidToTest->asteroid.size - 3) +
							(float) playerMaxStrength[P(playerNumber)] / 
							player[P(playerNumber)].spaceship->strength
						) * levelNumber;
				else
					scoreAssessed =
						abs(asteroidToTest->asteroid.size - 3) * levelNumber;
			//	printf("Strength: %d; factor: %f; score assessed: %d\n",player[P(playerNumber)].strength, (float) playerMaxStrength[P(playerNumber)]/player[P(playerNumber)].strength, scoreAssessed);	
				if(shipPtr->spaceship->invinciblityTimeToLive > 0)
					scoreAssessed /= 2;
				playerScore[P(playerNumber)] += scoreAssessed;

			}
			currentPtr = currentPtr->nextNode;
		}
	}

	
	return isHit;
}

/****************************************************
 * accelerateShip
 *  move a ship
 **************************/
void accelerateShip(Spaceship * player, float movementScaleFactor){
	float acceleration = 0.625; //magnitude to increase the direction (velocity) vector
	unsigned int counter, numberThrustSounds = 0;
	
	acceleration *= movementScaleFactor;
	if(player->spaceship->thrusting == true){
		for(counter = 0; counter < NUMBER_CHANNELS; counter++){
			if(channels[counter].soundID == THRUST_SOUND && 
					channels[counter].dataPosition != channels[counter].dataLength)
				numberThrustSounds++; //count how many times the thrust sound effect is playing
		}
		if(numberThrustSounds < numberOfPlayers)
			playSound(soundBuffer, THRUST_SOUND);

		player->direction.x += acceleration*(player->spaceship->strength/ (float) playerMaxStrength[P(player->spaceship->playerNumber)])*cos(player->spaceship->rotationAngle * PI/180);
		player->direction.y += acceleration*(player->spaceship->strength/ (float) playerMaxStrength[P(player->spaceship->playerNumber)])*sin(player->spaceship->rotationAngle * PI/180);
	}
}

/****************************************************
 * decelerateShip
 *  stops a ship, via the mystical reverse.
 **************************/
void decelerateShip(Spaceship * player, float movementScaleFactor){
	float acceleration = 0.5; //factor by which to decrease the direction (velocity) vector
	
	if( fabs(player->direction.x) > 0.1 || fabs(player->direction.y) > 0.1 ){
		acceleration *= movementScaleFactor;
		if(player->spaceship->decelerating == true){
			player->direction.x *= 1 - acceleration*(player->spaceship->strength/ (float) playerMaxStrength[P(player->spaceship->playerNumber)]);
			player->direction.y *= 1 - acceleration*(player->spaceship->strength/ (float) playerMaxStrength[P(player->spaceship->playerNumber)]);
		}
		playerScore[P(player->spaceship->playerNumber)] -= (int) floor( (24 + levelNumber)*movementScaleFactor);
	}
	else{ //the ship is barely moving, so just stop it.
		player->direction.x = 0;
		player->direction.y = 0;
	}
}


/****************************************************
 * rotateShip
 *  rotates a ship based on the value of its rotating flag
 **************************/
void rotateShip(Spaceship * player, float movementScaleFactor){
	float rotationValue = 10; //how many degrees to rotate the ship
	
	rotationValue = movementScaleFactor*rotationValue;
	if(player->spaceship->rotating != false){
		if(player->spaceship->rotationScaleFactor < 1) //to start the ship rotataing slowly for precision guidance
			player->spaceship->rotationScaleFactor = player->spaceship->rotationScaleFactor+0.20*movementScaleFactor;
		if(player->spaceship->rotating == -1) //counterclockwise
			player->spaceship->rotationAngle += rotationValue*player->spaceship->rotationScaleFactor;
		else //clockwise
			player->spaceship->rotationAngle -= rotationValue*player->spaceship->rotationScaleFactor;
			
		player->spaceship->rotationAngle = fmod(player->spaceship->rotationAngle, 360);
	}
}

/****************************************************
 * impactShip
 *  registers an impact; blows up a ship
 *  and slates it for rebuilding if necessary
 **************************/
void impactShip(Spaceship *ship, HitData impactData){
	int playerNumber = ship->spaceship->playerNumber;
	
	if(ship->spaceship->invinciblityTimeToLive <= 0){ //that is, not incinvible
		ship->spaceship->strength -= impactData.hitCode; //hitCode holds the strength of the impact
		if(ship->spaceship->strength < 0)
			ship->spaceship->strength = 0;
		if(impactData.hitCode != 0){
			playSound(soundBuffer, SHIELDS_SOUND);
			ship->spaceship->drawShieldsTimeToLive = 0 + FADEOUTS; //only draw a bubble for the fadeout time
		}
		if(ship->spaceship->strength <= 0){
			if(playerNumber > 0){ //this code is specific to actual players, not to *potential* enemies, if I get around to them. Enemies will have numbers less than 0
				//disable player if shields are gone
				playerIsEnabled[P(playerNumber)] = false;
				playerLives[P(playerNumber)]--;
				if(playerLives[P(playerNumber)] < 0){
					numberOfPlayersActive--;
					playerMaxLevel[P(playerNumber)] = levelNumber;
					if(numberOfPlayersActive <= 0)
						glutTimerFunc(3000, endGame, NULL); //end the game after 3 seconds; after all; no one is playing!
				}
				else{
					glutTimerFunc(0, resurrectShip, playerNumber); //rebuild after 3 seconds
				}
			}
			spawnKerblooie(
				KERBLOOIE_HUGE,
				ship->spaceship->shipSize/2,
				ship->spaceship->color,
				ship->position,
				ship->direction, //which has already been collided at the top of the function
				&kerblooieList );
		}
	}
	
}


/****************************************************
 * resurrectShip
 *  rebuilds a spaceship; always called by a timer function!
 **************************/
void resurrectShip(int playerNumber){
	static boolean firstCall = true; //this is only initialized upon variable creation
	if(firstCall == true){
		glutTimerFunc(3000, resurrectShip, playerNumber); //rebuild after 3 seconds
		firstCall = false;
		//Why this is necessary:
		//For some reason, calling glutTimerFunc (within the process of the idle function?) causes it to only time the first call. Subsequent calls execute immediately. Hence the "double call" here.
		//Inexplicably, the second timer call works just fine.
	}
	else{
		free(player[P(playerNumber)].spaceship); //since that pointer will be malloced on the next line
		player[P(playerNumber)] = spawnSpaceship(playerNumber);
		playerIsEnabled[P(playerNumber)] = true;
		firstCall = true;
	}
}



/****************************************************
 * spawnStarfield
 *  creates a pretty starfield for the background
 **************************/
void spawnStarfield() {
	unsigned int counter;
	for(counter = 0; counter < NUMBERSTARS; counter++){
		//one row: positionx, positiony, size, opacity
		starfield[counter][0] = rand()%FRAMEWIDTH; //x
		starfield[counter][1] = rand()%FRAMEHEIGHT; //y
		starfield[counter][2] = rand()%2 + 1; //point size: NOTE: this is unused. Reimplement with polygons instead of points to use it.
		starfield[counter][3] = floor( 255.0 * (float) rand()/ (float) RAND_MAX ); //opacity percentage
	//printf("%d %d %d %d\n",starfield[counter][0],starfield[counter][1],starfield[counter][2],starfield[counter][3]);
	}
}


/****************************************************
 * spawnSpaceship
 *  creates a spaceship
 **************************/
Spaceship spawnSpaceship(int number) {
	Spaceship alevin; //"alevin" is the term for a newly-hatched salmon. Hey, "spawned," "salmon," "alevin"--yeah...
	
	alevin.spaceship = malloc(sizeof(SpaceshipStruct));
	alevin.spaceship->playerNumber = number;
	alevin.spaceship->shipSize = 20;
	alevin.mass = 16;
	switch(number){
		case 1:
			alevin.spaceship->color[0] = player1Colors[0];
			alevin.spaceship->color[1] = player1Colors[1];
			alevin.spaceship->color[2] = player1Colors[2];
			break;
		case 2:
			alevin.spaceship->color[0] = player2Colors[0];
			alevin.spaceship->color[1] = player2Colors[1];
			alevin.spaceship->color[2] = player2Colors[2];
			break;
	}
	alevin.spaceship->rotationAngle = 0;
	alevin.position.x = FRAMEWIDTH  * number/(numberOfPlayers + 1);
	alevin.position.y = FRAMEHEIGHT * number/(numberOfPlayers + 1);
	alevin.direction.x = 0;
	alevin.direction.y = 0;
	alevin.spaceship->invinciblityTimeToLive = INVINCIBLE_LENGTH;
	alevin.spaceship->drawShieldsTimeToLive = alevin.spaceship->invinciblityTimeToLive + FADEOUTS/2;
	alevin.spaceship->strength = playerMaxStrength[P(number)];
	alevin.spaceship->thrusting = false;
	alevin.spaceship->drawThrustUntil = 0;
	alevin.spaceship->rotating = false;
	
	return alevin;
}


/****************************************************
 * spawnAsteroid
 *  creates an asteroid of specified size moving in a random direction
 **************************/
Object spawnAsteroid(int sizeIn, int positionX, int postionY, Vector hint) {
	Asteroid asteroid;
	int asteroidAngle;
	float magnitude;
	
	asteroid.asteroid.size = sizeIn;
	asteroid.asteroid.impacted.hitCode = false;
	asteroid.position.x = positionX;
	asteroid.position.y = postionY;
	
	asteroidAngle = rand()%360;
	//the Magnitude: a Random Number between 0 and 1, multiplied by a speed for the level
	
	if(sizeIn == ASTEROID_BIG){
		do{
			magnitude = (1+levelNumber/5)*((float) rand()/RAND_MAX + abs(sizeIn/sizeof(asteroidSizes)/sizeof(int)) - 1.5 );
		}while(magnitude == 0); //an asteroid must be moving!
	}
	else
		magnitude = magnitudeVector(hint) *(1 + fmod((float) rand()/RAND_MAX, 0.25) );
	asteroid.direction.x = magnitude * cos(asteroidAngle) + hint.x;
	asteroid.direction.y = magnitude * sin(asteroidAngle) + hint.y;
	asteroid.asteroid.color[0] = 64;
	asteroid.asteroid.color[1] = 0;
	asteroid.asteroid.color[2] = 128;
	asteroid.mass = (sizeIn+1)*12;
	
	return asteroid;
}


/****************************************************
 * spawnKerblooie
 *  creates an explosion of specified KERBLOOIE_SIZE at specified point, with specified directional hint
 *	put particles in specified list
 **************************/
void spawnKerblooie(int sizeIn, int radius, GLubyte *colorHint, Coordinate position, Vector hint, NodePtr *list) {
	Kerblooie particle;
	unsigned int counter, subcounter;
	int kerblooieSize;
	
	int particleAngle;
	float magnitude;
	
	playSound(soundBuffer, sizeIn + EXPLOSION_SMALL_SOUND); //since kerblooie sizes range from 0 (small) to 3 (huge), and explosion sound indexes increment the same way
	kerblooieSize = kerblooieSizes[sizeIn];
	particle.mass = 0; //hardly worth setting
	
	for(counter = 0; counter < kerblooieSize; counter++){
		particleAngle = counter*4*2*PI/kerblooieSize;

		//Basic position function: r = kerblooieRadius*(theta/numberSpirals*2*PI), 0 <= theta <= numberSpirals*2*PI
		//Converted to cartesian ( x = r*cos(theta) ):
		particle.position.x = radius*particleAngle/(4*2*PI)*cos(particleAngle) + position.x;
		particle.position.y = radius*particleAngle/(4*2*PI)*sin(particleAngle) + position.y;

		//the Magnitude: a Random Number between 0 and 1, multiplied by a speed. 16 is the max speed; only achieveable on the biggest explosion
		do{
			magnitude = (float) rand()/RAND_MAX * ((float) sizeIn/KERBLOOIE_HUGE +1); //cast random to float beforehand; Integer/(float) Integer does not do a float division.
			//printf("Magnitude: %f; Rand: %d; RandMax = %d\n", magnitude, random, RAND_MAX);
		}while(magnitude == 0); //an asteroid must be moving!
		
		particle.direction.x = magnitude * cos(particleAngle) + hint.x;
		particle.direction.y = magnitude * sin(particleAngle) + hint.y;
		
		particle.kerblooie.lifespan = (rand()%2000 + 250) * ((float) sizeIn/KERBLOOIE_HUGE +1); // (0.25 < t < 2 seconds)*explosionSize
		particle.kerblooie.timeToLive = particle.kerblooie.lifespan;
		particle.kerblooie.size = (float) rand()/RAND_MAX * ( rand()%(sizeIn+1) ) + 2.25;
		for(subcounter = 0; subcounter < 3; subcounter++)
			particle.kerblooie.color[subcounter] = colorHint[subcounter] + rand()%(256-colorHint[subcounter])*pow(-1,rand()); //takes hinted color and gives it some variation
		particle.kerblooie.color[3] = 255;
				
		insertNode(list, particle);
	}	
}


/****************************************************
 * spawnCrystal
 *  creates a crystal of specified CRYSTAL_SPECIMEN at specified point
 **************************/
void spawnCrystal(int specimen, Coordinate position, NodePtr *list) {
	Crystal newCrystal;
	int crystalAngle; //the direction at which it goes zooming off
	float magnitude;
	
	playSound(soundBuffer, RELEASE_CRYSTAL_SOUND);
	newCrystal.crystal = malloc(sizeof(CrystalStruct));
	newCrystal.position = position;
	newCrystal.mass = 2;
	
	crystalAngle = rand()%360;
	//the Magnitude: level Number plus a random adjustment.
	do{
		magnitude = levelNumber/4.0;
		//printf("Magnitude: %f; Rand: %d; RandMax = %d\n", magnitude, random, RAND_MAX);
	}while(magnitude == 0); //a crystal must be moving!
	
	newCrystal.direction.x = magnitude * cos(crystalAngle);
	newCrystal.direction.y = magnitude * sin(crystalAngle);
	
	newCrystal.crystal->crystalFlavor = specimen;
	newCrystal.crystal->timeToLive = 10000; //10 seconds
	newCrystal.crystal->rotationIncrement = (float) rand()/RAND_MAX + 30/FRAMES_PER_SECOND;
	if(specimen == CRYSTAL_SHIELDPOWER){
		newCrystal.crystal->radius = 8;
		newCrystal.crystal->thickness = 1;
	//	newCrystal.crystal->radius = 128;
	//	newCrystal.crystal->thickness = 16;
		newCrystal.crystal->color[0] = 0;
		newCrystal.crystal->color[1] = 128;
		newCrystal.crystal->color[2] = 128;
		newCrystal.crystal->color[3] = 192;
	}
	newCrystal.crystal->status.hitCode = CRYSTAL_NORMAL;
	newCrystal.crystal->status.mass = newCrystal.mass;
	newCrystal.crystal->status.direction = newCrystal.direction;
	
	insertNode(list, newCrystal);
}


/****************************************************
 * startLevel
 *  initiates a level
 **************************/
void startLevel(int level) {
	static boolean firstCall = true; //this is only initialized upon variable creation
	unsigned int numberUnsmallAsteroids = 0;
	unsigned int totalNumberLargeAsteroids; //this will be the initial number of asteroids on the screen
	Vector zeroVector;
	int numberCrystalsToRelease[CRYSTAL_NUMBER_OF_SPECIMENS] = {0}; //Initialize to zero
	unsigned int counter, scheduleCounter;
	
	#ifdef RANDOM_ON
		srand( time(NULL) );
	#endif
	zeroVector.x = 0;
	zeroVector.y = 0;
	
	if(firstCall == true && startingNewLevel == true){ //don't do this silly jerry-rig unless the method was actually called during gameplay
		glutTimerFunc(4000, startLevel, level); //rebuild after 4 seconds
		firstCall = false;
		//Why this is necessary:
		//For some reason, calling glutTimerFunc (within the process of the idle function?) causes it to only time the first call. Subsequent calls execute immediately. Hence the "double call" here.
		//Inexplicably, the second timer call works just fine.
	}
	else{
		//the actual startLevel code!

		//reset players to default positions
		for(counter = 1; counter <= numberOfPlayers; counter++){
			if(playerIsEnabled[P(counter)] == true)
				playerScore[P(counter)] += 48 * (levelNumber); //the global level number is still the number of the old level just ended
	
			if(playerIsEnabled[P(counter)] == true){
				//stop the spaceship
				player[P(counter)].spaceship->rotationAngle = 0;
				player[P(counter)].position.x = FRAMEWIDTH  * counter/(numberOfPlayers + 1);
				player[P(counter)].position.y = FRAMEHEIGHT * counter/(numberOfPlayers + 1);
				player[P(counter)].direction.x = 0;
				player[P(counter)].direction.y = 0;
				player[P(counter)].spaceship->invinciblityTimeToLive = INVINCIBLE_LENGTH;
				player[P(counter)].spaceship->drawShieldsTimeToLive = player[P(counter)].spaceship->invinciblityTimeToLive + FADEOUTS/2;
				player[P(counter)].spaceship->thrusting = 0;
			}
		
		}
		
		//Wipe lists
		free(shotList);
		shotList = createList();
		free(crystalList);
		crystalList = createList();
		
		//Change background
		spawnStarfield();
		
		//create Asteroids and asteroid schedule
		/*
		 *	Only unsmall asteroids (big and medium) are "scheduled"--that is,
		 *	there is a preordained total number of unsmall asteroids that can
		 *	be produced, and the order in which they appear is also preordained.
		 *	(Small asteroids are left to chance.) This system is used to more evenly
		 *	distribute crystals throughout a level.
		 */
		totalNumberLargeAsteroids = level+2;
		numberUnsmallAsteroids = totalNumberLargeAsteroids; //start out with at least that many unsmalls
		free(asteroidReleaseSchedule);
		asteroidReleaseSchedule = calloc(totalNumberLargeAsteroids, sizeof(int) ); //will hold the number of smaller asteroids to be produced from each *large* one
		for(counter = 0; counter < totalNumberLargeAsteroids; counter++){ //spawn large asteroids and ordain their number of divisions
			insertNode( &asteroidList,
				spawnAsteroid(ASTEROID_BIG,rand() % FRAMEWIDTH, rand() % FRAMEHEIGHT , zeroVector) );
			asteroidReleaseSchedule[counter] = rand()%2+2; //either 2 or three new asteroids to be released
			numberUnsmallAsteroids += asteroidReleaseSchedule[counter]; //the number of unsmalls has now been increased by the random number just generated
		}
		unsmallAsteroidsDestroyedToDate = 0;
		bigAsteroidsDestroyedToDate = 0; //keep track of these as the array index for the schedule!
		//printf("\n\n\nUnsmall: %d\n", numberUnsmallAsteroids);
		
		//create Crystal schedule
		#ifdef DEBUG_CRYSTALS
			printf("\n\n");
		#endif
		numberCrystalsToRelease[CRYSTAL_SHIELDPOWER] = level*3/4 + ( rand()%level )/2;
		for(counter = 0; counter < CRYSTAL_NUMBER_OF_SPECIMENS; counter++){
			free(crystalReleaseSchedule[counter]);
			crystalReleaseSchedule[counter] = calloc(numberCrystalsToRelease[counter],sizeof(int));
			for(scheduleCounter = 0; scheduleCounter < numberCrystalsToRelease[counter]; scheduleCounter++){
				do{
					crystalReleaseSchedule[counter][scheduleCounter] = rand()%numberUnsmallAsteroids; //try a random number somewhere amongst the unsmall explosions
				}while( scheduleCounter * numberUnsmallAsteroids / numberCrystalsToRelease[CRYSTAL_SHIELDPOWER] >= crystalReleaseSchedule[counter][scheduleCounter] || //these two lines specify the bracket conditions (lower and upper), which are vital to ensuring the crystals are *evenly* distributed
						crystalReleaseSchedule[counter][scheduleCounter] > (scheduleCounter+1)*numberUnsmallAsteroids/numberCrystalsToRelease[CRYSTAL_SHIELDPOWER] //value must be within the desired bracket
						|| (
							(scheduleCounter > 0) ? //conditional operator: if(scheduleCounter>0) then (blah) else true
								crystalReleaseSchedule[counter][scheduleCounter] <= crystalReleaseSchedule[counter][scheduleCounter-1] //make sure that we *keep looping* if the computer hasn't generated a number greater than the previous one it came up with
							:
								false //this only happens if we are still on the very first scheduled crystal, on which we don't need to consider the previous line
							)
					); //loop while conditions are wrong
				if(numberCrystalsToRelease[counter] == 0)
					crystalReleaseSchedule[counter][scheduleCounter] = -1; //when there are no crystals to create, one schedule slot gets calloced and filled with 0 anyway. Fill it with a clearly invalid number instead.
				#ifdef DEBUG_CRYSTALS
					printf("%d, ",crystalReleaseSchedule[counter][scheduleCounter]);
				#endif
			}
			crystalsReleased[counter] = 0;
			#ifdef DEBUG_CRYSTALS
				printf("\nCyrstals (%d) to be released: %d\n", counter, numberCrystalsToRelease[counter]);
			#endif
		}
		
 		startingNewLevel = false; //the new level has begun
		levelNumber = level;

		firstCall = true;
	}


}


/****************************************************
 * endGame
 **************************/
void endGame(int needlessParameter) {
	static boolean firstCall = true; //this is only initialized upon variable creation
	if(firstCall == true){
		glutTimerFunc(3000, endGame, needlessParameter); //rebuild after 3 seconds
		firstCall = false;
		//Why this is necessary:
		//For some reason, calling glutTimerFunc (within the process of the idle function?) causes it to only time the first call. Subsequent calls execute immediately. Hence the "double call" here.
		//Inexplicably, the second timer call works just fine.
	}
	else{
		//the actual endGame code!
		titleInit();
		newHighScore(1); //create highscore for player1. Player2's is called afterwards.
	
		glutPostRedisplay();
		firstCall = true;
	}
}

/****************************************************
 * handleKeyboard
 *  accepts regular keyboard input
 **************************/
void handleKeyboard(unsigned char keypressed, int x, int y){
	handleKey(keypressed);
}

/****************************************************
 * handleSpecialKeyboard
 *  accepts special keys, i.e. arrows, etc.
 **************************/
void handleSpecialKeyboard(int keypressed, int x, int y){
	handleKey(-keypressed); //by convention in my program, special keys are negative
}


/****************************************************
 * handleKey
 *  accepts regular keyboard input
 **************************/
void handleKey(int keypressed){
	unsigned int counter;
	
	//Cheat Codes!
	if(glutGetModifiers() == GLUT_ACTIVE_ALT){
		switch(keypressed){
			case -GLUT_KEY_F1:
				playerLives[P(1)]++;
				break;
			case -GLUT_KEY_F2:
				playerLives[P(2)]++;
				break;
			case -GLUT_KEY_F5:	
				levelNumber++;
				break;
			case -GLUT_KEY_F6:	
				levelNumber--;
				break;
			
		} //end switch
	}//end if

	//pause on escape keypress
	if(keypressed == 27 && paused == false){
		glutSpecialFunc(NULL);
		glutIdleFunc(NULL);
		
		paused = true;
		glutPostRedisplay();
	}
	else if(keypressed == 27 && paused == true){ //*unpause only* on escape key
		glutKeyboardUpFunc(handleKeyboardUp);
		glutSpecialFunc(handleSpecialKeyboard);
		glutSpecialUpFunc(handleSpecialKeyboardUp);
		glutIdleFunc(idle);
		
		paused = false;
	}
	else if(paused == false){ //don't do key actions if paused
		for(counter = 1; counter <= numberOfPlayers; counter++){
	
			if(playerIsEnabled[P(counter)] == true){
				if( FIRE(P(counter)) == keypressed )
					fireShot(&player[P(counter)]);
				else if( ROTATE_CCW(P(counter)) == keypressed ){
					player[P(counter)].spaceship->rotating = -1; //Counter-clockwise
					player[P(counter)].spaceship->rotationScaleFactor = 0; //this value is incremented up to one as the ship continues to rotate. This allows precision alignment at first.					
				}
				else if( ROTATE_CW(P(counter)) == keypressed){
					player[P(counter)].spaceship->rotating = 1; //Clockwise
					player[P(counter)].spaceship->rotationScaleFactor = 0;					
				}
				else if( THRUST(P(counter)) == keypressed )
					player[P(counter)].spaceship->thrusting = true;
				else if( REVERSE(P(counter)) == keypressed )
					player[P(counter)].spaceship->decelerating = true;
			}
		}//end for
	}
}


/****************************************************
 * handleKeyboardUp
 *  accepts regular keyboard input
 **************************/
void handleKeyboardUp(unsigned char keypressed, int x, int y){
	handleKeyUp(keypressed);
}

/****************************************************
 * handleSpecialKeyboardUp
 *  accepts special keys, i.e. arrows, etc.
 **************************/
void handleSpecialKeyboardUp(int keypressed, int x, int y){
	handleKeyUp(-keypressed); //by convention in my program, special keys are negative
}


/****************************************************
 * handleKeyUp
 *  when a key is released
 **************************/
void handleKeyUp(int keyreleased){
	unsigned int counter;
	
	for(counter = 1; counter <= numberOfPlayers; counter++){

		if(playerIsEnabled[P(counter)] == true){
			if( ROTATE_CCW(P(counter)) == keyreleased)
				player[P(counter)].spaceship->rotating = false; //Counter-clockwise
			else if( ROTATE_CW(P(counter)) == keyreleased)
				player[P(counter)].spaceship->rotating = false;
			else if( THRUST(P(counter)) == keyreleased){
				player[P(counter)].spaceship->thrusting = false;
				player[P(counter)].spaceship->drawThrustUntil = glutGet( GLUT_ELAPSED_TIME ) + FADEOUTS;
				playSound(soundBuffer, STOP_THRUSTING_SOUND);
			}
			else if( REVERSE(P(counter)) == keyreleased )
				player[P(counter)].spaceship->decelerating = false;
		}
	
	}//end for
}


/****************************************************
 * List Fuctions
 *  handle a linked list
 **************************/

/*createList--returns a NodePtr to a new blank doubly-linked list*/
NodePtr createList(){
	NodePtr header,trailer;
	
	//Initialized Basic Doubly-Linked List Nodes
	header = malloc( sizeof(Node) );
	trailer = malloc( sizeof(Node) );

	header->previousNode = NULL;
	//header->object = NULL;
	header->nextNode = trailer;
	trailer->previousNode = header;
	//tailer->object = NULL;
	trailer->nextNode = NULL;
	
	return header;
}


/*Insert--Places an object at the point of startPtr*/
void insertNode( NodePtr *startPtr, Object valueIn ){
	NodePtr newPtr, nextPtr;
	newPtr = malloc( sizeof(Node) );
	
	if(newPtr != NULL ){
		newPtr->object = valueIn;
		if(*startPtr == NULL){
			exit(1); //it is a crash to insert an item into a null list
		}
	//	else if( (*startPtr)->previousNode == NULL ){ //Inserting at head of list
		else{
			nextPtr = (*startPtr)->nextNode;
			
			newPtr->previousNode = *startPtr;
			newPtr->nextNode = nextPtr;
			(*startPtr)->nextNode = newPtr;
			nextPtr->previousNode = newPtr;
		}
	/*	else{
			newPtr->previousNode = (*startPtr)->previousNode;
			newPtr->nextNode = *startPtr;
			prevPtr = (*startPtr)->previousNode;
			prevPtr->nextNode = newPtr;
			(*startPtr)->previousNode = newPtr;
		}
	*/
	}
	else{
		endGame(NULL);
	}
}

/*Remove--Wipes the specified node out of the list*/
Object removeNode( NodePtr *nodePtr/*, Object valueIn */){
	Object toReturn = (*nodePtr)->object;
	NodePtr prevPtr, nextPtr, tempPtr;
	
	if( !isEmpty(*nodePtr) ){
		tempPtr = *nodePtr;
		nextPtr = (*nodePtr)->nextNode;
		prevPtr = (*nodePtr)->previousNode;
		
		if(nextPtr != NULL) //not at tail of list
			nextPtr->previousNode = (*nodePtr)->previousNode;
		if(prevPtr != NULL) //not at head of list
			prevPtr->nextNode = (*nodePtr)->nextNode;
		//*nodePtr = nextPtr;
		free( tempPtr );
	}
	return toReturn;
}

/*isEmpty--Tests for the Empty List*/
boolean isEmpty( NodePtr nodePtr ){
	if(nodePtr == NULL || (nodePtr->previousNode == NULL && nodePtr->nextNode->nextNode == NULL) )
		return true;
	else
		return false;
}